"""
Analyzer for a UXsim simulation result.
This module is automatically loaded when you import the `uxsim` module.
"""

import numpy as np
import glob, os, csv, time
import pandas as pd
from collections import defaultdict as ddict
from scipy.sparse.csgraph import floyd_warshall
from .utils import *

class Analyzer:
    """
    Class for analyzing and visualizing a simulation result.
    """

    def __init__(s, W, font_pillow=None, font_matplotlib=None):
        """
        Create a result analysis object.

        Parameters
        ----------
        W : object
            The world to which this belongs.
        font_pillow : str, optional
            The path to the font file for Pillow. If not provided, the default font for English and Japanese is used.
        font_matplotlib : str, optional
            The font name for Matplotlib. If not provided, the default font for English and Japanese is used.
        """
        s.W = W

        # os.makedirs(f"out{s.W.name}", exist_ok=True)

        #基礎統計量
        s.average_speed = 0
        s.average_speed_count = 0
        s.trip_completed = 0
        s.trip_all = 0
        s.total_travel_time = 0
        s.average_travel_time = 0

        #フラグ
        s.flag_edie_state_computed = 0
        s.flag_trajectory_computed = 0
        s.flag_pandas_convert = 0
        s.flag_od_analysis = 0

    def basic_analysis(s):
        """
        Analyze basic stats.
        """
        df = s.W.analyzer.od_to_pandas()

        s.trip_completed = np.sum(df["completed_trips"])
        s.trip_all = np.sum(df["total_trips"])
        
        s.total_distance_traveled = np.sum(df["average_distance_traveled_per_veh"]*df["total_trips"])

        if s.trip_completed:
            s.total_travel_time = np.sum(df["completed_trips"]*df["average_travel_time"])
            s.average_travel_time = s.total_travel_time/s.trip_completed
            s.total_delay = np.sum(df["completed_trips"]*(df["average_travel_time"]-df["free_travel_time"]))
            s.average_delay = s.total_delay/s.trip_completed
        else:
            s.total_travel_time = -1
            s.average_travel_time = -1
            s.total_delay = -1
            s.average_delay = -1


    def od_analysis(s):
        """
        Analyze OD-specific stats: number of trips, number of completed trips, free-flow travel time, average travel time, its std, total distance traveled.
        """
        if s.flag_od_analysis:
            return 0
        else:
            s.flag_od_analysis = 1

        s.od_trips = ddict(lambda: 0)
        s.od_trips_comp = ddict(lambda: 0)
        s.od_tt_free = ddict(lambda: 0)
        s.od_tt = ddict(lambda: [])
        s.od_tt_ave = ddict(lambda: 0)
        s.od_tt_std = ddict(lambda: 0)
        s.od_dist = ddict(lambda: [])
        s.od_dist_total = ddict(lambda: 0)
        s.od_dist_ave = ddict(lambda: 0)
        s.od_dist_std = ddict(lambda: 0)
        s.od_dist_min = ddict(lambda: 0)
        dn = s.W.DELTAN

        #自由旅行時間と最短距離
        adj_mat_time = np.zeros([len(s.W.NODES), len(s.W.NODES)])
        adj_mat_dist = np.zeros([len(s.W.NODES), len(s.W.NODES)])
        for link in s.W.LINKS:
            i = link.start_node.id
            j = link.end_node.id
            if s.W.ADJ_MAT[i,j]:
                adj_mat_time[i,j] = link.length/link.u
                adj_mat_dist[i,j] = link.length
                if link.capacity_in == 0: #流入禁止の場合は通行不可
                    adj_mat_time[i,j] = np.inf
                    adj_mat_dist[i,j] = np.inf
            else:
                adj_mat_time[i,j] = np.inf
                adj_mat_dist[i,j] = np.inf
        dist_time = floyd_warshall(adj_mat_time)
        dist_space = floyd_warshall(adj_mat_dist)

        for veh in s.W.VEHICLES.values():
            o = veh.orig
            d = veh.dest
            if d != None:
                s.od_trips[o,d] += dn

                veh_links = [rec[1] for rec in veh.log_t_link if hasattr(rec[1], "length")]
                veh_dist_traveled = sum([l.length for l in veh_links])
                if veh.state == "run":
                    veh_dist_traveled += veh.x
                veh.distance_traveled = veh_dist_traveled
                s.od_dist[o,d].append(veh.distance_traveled)

                if veh.travel_time != -1:
                    s.od_trips_comp[o,d] += dn
                    s.od_tt[o,d].append(veh.travel_time)
        for o,d in s.od_tt.keys():
            s.od_tt_ave[o,d] = np.average(s.od_tt[o,d])
            s.od_tt_std[o,d] = np.std(s.od_tt[o,d])
            s.od_tt_free[o,d] = dist_time[o.id, d.id]
            s.od_dist_total[o,d] = np.sum(s.od_dist[o,d])
            s.od_dist_min[o,d] = dist_space[o.id, d.id]
            s.od_dist_ave[o,d] = np.average(s.od_dist[o,d])
            s.od_dist_std[o,d] = np.std(s.od_dist[o,d])

    def link_analysis_coarse(s):
        """
        Analyze link-level coarse stats: traffic volume, remaining vehicles, free-flow travel time, average travel time, its std.
        """
        s.linkc_volume = ddict(lambda:0)
        s.linkc_tt_free = ddict(lambda:0)
        s.linkc_tt_ave = ddict(lambda:-1)
        s.linkc_tt_std = ddict(lambda:-1)
        s.linkc_remain = ddict(lambda:0)

        for l in s.W.LINKS:
            s.linkc_volume[l] = l.cum_departure[-1]
            s.linkc_remain[l] = l.cum_arrival[-1]-l.cum_departure[-1]
            s.linkc_tt_free[l] = l.length/l.u
            if s.linkc_volume[l]:
                s.linkc_tt_ave[l] = np.average([t for t in l.traveltime_actual if t>0])
                s.linkc_tt_std[l] = np.std([t for t in l.traveltime_actual if t>0])

    def compute_accurate_traj(s):
        """
        Generate more complete vehicle trajectories for each link by extrapolating recorded trajectories. It is assumed that vehicles are in free-flow travel at the end of the link.
        """
        if s.W.vehicle_logging_timestep_interval != 1:
            warnings.warn("vehicle_logging_timestep_interval is not 1. The trajectories are not exactly accurate.", LoggingWarning)

        if s.flag_trajectory_computed:
            return 0
        else:
            s.flag_trajectory_computed = 1

        for veh in s.W.VEHICLES.values():
            l_old = None
            for i in lange(veh.log_t):
                if veh.log_link[i] != -1:
                    l = s.W.get_link(veh.log_link[i])
                    if l_old != l:
                        l.tss.append([])
                        l.xss.append([])
                        l.ls.append(veh.log_lane[i])
                        l.cs.append(veh.color)
                        l.names.append(veh.name)

                    l_old = l
                    l.tss[-1].append(veh.log_t[i])
                    l.xss[-1].append(veh.log_x[i])

        for l in s.W.LINKS:
            #端部を外挿
            for i in lange(l.xss):
                if len(l.xss[i]):
                    if l.xss[i][0] != 0:
                        x_remain = l.xss[i][0]
                        if x_remain/l.u > s.W.DELTAT*0.01:
                            l.xss[i].insert(0, 0)
                            l.tss[i].insert(0, l.tss[i][0]-x_remain/l.u)
                    if l.length-l.u*s.W.DELTAT <= l.xss[i][-1] < l.length:
                        x_remain = l.length-l.xss[i][-1]
                        if x_remain/l.u > s.W.DELTAT*0.01:
                            l.xss[i].append(l.length)
                            l.tss[i].append(l.tss[i][-1]+x_remain/l.u)

    def compute_edie_state(s):
        """
        Compute Edie's traffic state for each link.
        """
        if s.flag_edie_state_computed:
            return 0
        else:
            s.flag_edie_state_computed = 1

        s.compute_accurate_traj()
        for l in s.W.LINKS:
            DELTAX = l.edie_dx
            DELTATE = l.edie_dt
            MAXX = l.length
            MAXT = s.W.TMAX

            dt = DELTATE
            dx = DELTAX
            tn = [[ddict(lambda: 0) for i in range(int(MAXX/DELTAX))] for j in range(int(MAXT/DELTATE))]
            dn = [[ddict(lambda: 0) for i in range(int(MAXX/DELTAX))] for j in range(int(MAXT/DELTATE))]

            l.k_mat = np.zeros([int(MAXT/DELTATE), int(MAXX/DELTAX)])
            l.q_mat = np.zeros([int(MAXT/DELTATE), int(MAXX/DELTAX)])
            l.v_mat = np.zeros([int(MAXT/DELTATE), int(MAXX/DELTAX)])

            for v in lange(l.xss):
                for i in lange(l.xss[v][:-1]):
                    i0 = l.names[v]
                    x0 = l.xss[v][i]
                    x1 = l.xss[v][i+1]
                    t0 = l.tss[v][i]
                    t1 = l.tss[v][i+1]
                    if t1-t0 != 0:
                        v0 = (x1-x0)/(t1-t0)
                    else:
                        #compute_accurate_traj()の外挿で極稀にt1=t0になったのでエラー回避（もう起きないはずだが念のため）
                        v0 = 0

                    tt = int(t0//dt)
                    xx = int(x0//dx)

                    if v0 > 0:
                        #残り
                        xl0 = dx-x0%dx
                        xl1 = x1%dx
                        tl0 = xl0/v0
                        tl1 = xl1/v0

                        if tt < int(MAXT/DELTATE) and xx < int(MAXX/DELTAX):
                            if xx == x1//dx:
                                #(x,t)
                                dn[tt][xx][i0] += x1-x0
                                tn[tt][xx][i0] += t1-t0
                            else:
                                #(x+n, t)
                                jj = int(x1//dx-xx+1)
                                for j in range(jj):
                                    if xx+j < int(MAXX/DELTAX):
                                        if j == 0:
                                            dn[tt][xx+j][i0] += xl0
                                            tn[tt][xx+j][i0] += tl0
                                        elif j == jj-1:
                                            dn[tt][xx+j][i0] += xl1
                                            tn[tt][xx+j][i0] += tl1
                                        else:
                                            dn[tt][xx+j][i0] += dx
                                            tn[tt][xx+j][i0] += dx/v0
                    else:
                        if tt < int(MAXT/DELTATE) and xx < int(MAXX/DELTAX):
                            dn[tt][xx][i0] += 0
                            tn[tt][xx][i0] += t1-t0

            for i in lange(tn):
                for j in lange(tn[i]):
                    t = list(tn[i][j].values())*s.W.DELTAN
                    d = list(dn[i][j].values())*s.W.DELTAN
                    l.tn_mat[i,j] = sum(t)
                    l.dn_mat[i,j] = sum(d)
                    l.k_mat[i,j] = l.tn_mat[i,j]/DELTATE/DELTAX
                    l.q_mat[i,j] = l.dn_mat[i,j]/DELTATE/DELTAX
            with np.errstate(invalid="ignore"):
                l.v_mat = l.q_mat/l.k_mat
            l.v_mat = np.nan_to_num(l.v_mat, nan=l.u)

    @catch_exceptions_and_warn()
    def print_simple_stats(s, force_print=False):
        """
        Prints basic statistics of simulation result.

        Parameters
        ----------
        force_print : bool, optional
            print the stats regardless of the value of `print_mode`
        """
        s.W.print("results:")
        s.W.print(f" average speed:\t {s.average_speed:.1f} m/s")
        s.W.print(" number of completed trips:\t", s.trip_completed, "/", s.trip_all)
        #s.W.print(" number of completed trips:\t", s.trip_completed, "/", len(s.W.VEHICLES)*s.W.DELTAN)
        if s.trip_completed > 0:
            s.W.print(f" total travel time:\t\t {s.total_travel_time:.1f} s")
            s.W.print(f" average travel time of trips:\t {s.average_travel_time:.1f} s")
            s.W.print(f" average delay of trips:\t {s.average_delay:.1f} s")
            s.W.print(f" delay ratio:\t\t\t {s.average_delay/s.average_travel_time:.3f}")
        s.W.print(f" total distance traveled:\t {s.total_distance_traveled:.1f} m")

        if force_print == 1 and s.W.print_mode == 0:
            print("results:")
            print(f" average speed:\t {s.average_speed:.1f} m/s")
            print(" number of completed trips:\t", s.trip_completed, "/", s.trip_all)
            #print(" number of completed trips:\t", s.trip_completed, "/", len(s.W.VEHICLES)*s.W.DELTAN)
            if s.trip_completed > 0:
                print(f" total travel time:\t\t {s.total_travel_time:.1f} s")
                print(f" average travel time of trips:\t {s.average_travel_time:.1f} s")
                print(f" average delay of trips:\t {s.average_delay:.1f} s")
                print(f" delay ratio:\t\t\t {s.average_delay/s.average_travel_time:.3f}")
            print(f" total distance traveled:\t {s.total_distance_traveled:.1f} m")


    def comp_route_travel_time(s, t, route):
        pass


    @catch_exceptions_and_warn()
    def show_simulation_progress(s):
        """
        Print simulation progress.
        """
        if s.W.print_mode:
            vehs = [l.density*l.length for l in s.W.LINKS]
            sum_vehs = sum(vehs)

            vs = [l.density*l.length*l.speed for l in s.W.LINKS]
            if sum_vehs > 0:
                avev = sum(vs)/sum_vehs
            else:
                avev = 0

            print(f"{s.W.TIME:>8.0f} s| {sum_vehs:>8.0f} vehs|  {avev:>4.1f} m/s| {time.time()-s.W.sim_start_time:8.2f} s", flush=True)


    def compute_mfd(s, links=None):
        """
        Compute network average flow and density for MFD.
        """
        s.compute_edie_state()
        if links == None:
            links = s.W.LINKS
        links = [s.W.get_link(link) for link in links]
        links = frozenset(links)


        for i in range(len(s.W.Q_AREA[links])):
            tn = sum([l.tn_mat[i,:].sum() for l in s.W.LINKS if l in links])
            dn = sum([l.dn_mat[i,:].sum() for l in s.W.LINKS if l in links])
            an = sum([l.length*s.W.EULAR_DT for l in s.W.LINKS if l in links])
            s.W.K_AREA[links][i] = tn/an
            s.W.Q_AREA[links][i] = dn/an


    def vehicles_to_pandas(s):
        """
        Compute the detailed vehicle travel logs and return as a pandas DataFrame.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing the travel logs of vehicles, with the columns:
            
            - 'name': the name of the vehicle (platoon).
            - 'dn': the platoon size.
            - 'orig': the origin node of the vehicle's trip.
            - 'dest': the destination node of the vehicle's trip.
            - 't': the timestep.
            - 'link': the link the vehicle is on (or relevant status).
            - 'x': the position of the vehicle on the link.
            - 's': the spacing of the vehicle.
            - 'v': the speed of the vehicle.
        """
        if s.W.vehicle_logging_timestep_interval != 1:
            warnings.warn("vehicle_logging_timestep_interval is not 1. The output data is not exactly accurate.", LoggingWarning)

        if s.flag_pandas_convert == 0:
            out = [["name", "dn", "orig", "dest", "t", "link", "x", "s", "v"]]
            for veh in s.W.VEHICLES.values():
                for i in range(len(veh.log_t)):
                    if veh.log_state[i] in ("wait", "run", "end", "abort"):
                        if veh.log_link[i] != -1:
                            linkname = veh.log_link[i].name
                        else:
                            if veh.log_state[i] == "wait":
                                linkname = "waiting_at_origin_node"
                            elif veh.log_state[i] == "abort":
                                linkname = "trip_aborted"
                            else:
                                linkname = "trip_end"
                        veh_dest_name = None
                        if veh.dest != None:
                            veh_dest_name = veh.dest.name
                        out.append([veh.name, s.W.DELTAN, veh.orig.name, veh_dest_name, veh.log_t[i], linkname, veh.log_x[i], veh.log_s[i], veh.log_v[i]])
            s.df_vehicles = pd.DataFrame(out[1:], columns=out[0])

            s.flag_pandas_convert = 1
        return s.df_vehicles

    def log_vehicles_to_pandas(s):
        """
        Same to `vehicles_to_pandas`, just for backward compatibility.
        """
        return s.vehicles_to_pandas()

    def vehicle_trip_to_pandas(s):
        """
        Compute the vehicle trip summary and return as a pandas DataFrame.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing the trip summary of the vehicle trip logs, with the columns:
            
            - 'name': the name of the vehicle (platoon).
            - 'orig': the origin node of the vehicle's trip.
            - 'dest': the destination node of the vehicle's trip.
            - 'departure_time': the departure time of the vehicle.
            - 'final_state': the final state of the vehicle.
            - 'travel_time': the travel time of the vehicle.
            - 'average_speed': the average speed of the vehicle.
            - 'distance_traveled': the distance traveled by the vehicle.
        """
        out = [["name", "orig", "dest", "departure_time", "final_state", "travel_time", "average_speed", "distance_traveled"]]
        for veh in s.W.VEHICLES.values():
            veh_dest_name = veh.dest.name if veh.dest != None else None
            veh_state = veh.log_state[-1]
            veh_ave_speed = np.average([v for v in veh.log_v if v != -1])
            veh_dist_traveled = veh.distance_traveled

            out.append([veh.name, veh.orig.name, veh_dest_name, veh.departure_time*s.W.DELTAT, veh_state, veh.travel_time, veh_ave_speed, veh_dist_traveled])
        
        s.df_vehicle_trip = pd.DataFrame(out[1:], columns=out[0])
        return s.df_vehicle_trip

    def gps_like_log_to_pandas(s):
        """
        Generate GPS-like log (x and y in the coordinate system used for Node) of vehicles and return as a pandas DataFrame.

        Returns
        -------
        pd.DataFrame
        """
        out = [["name", "t", "x", "y", "v"]]
        for veh in s.W.VEHICLES.values():
            for i,t in enumerate(veh.log_t):
                x, y = veh.get_xy_coords(t)
                if (x, y) == (-1, -1):
                    continue
                v = veh.log_v[i]
                out.append([veh.name, t, x, y, v])
        s.df_gps_like_log = pd.DataFrame(out[1:], columns=out[0])
        return s.df_gps_like_log

    def basic_to_pandas(s):
        """
        Comutes the basic stats and return as a pandas DataFrame.

        Returns
        -------
        pd.DataFrame
        """
        out = [["total_trips", "completed_trips", "total_travel_time", "average_travel_time", "total_delay", "average_delay"], [s.trip_all, s.trip_completed, s.total_travel_time, s.average_travel_time, s.total_delay, s.average_delay]]

        s.df_basic = pd.DataFrame(out[1:], columns=out[0])
        return s.df_basic

    def od_to_pandas(s):
        """
        Compute the OD-specific stats and return as a pandas DataFrame.

        Returns
        -------
        pd.DataFrame
        """

        s.od_analysis()

        out = [["orig", "dest", "total_trips", "completed_trips", "free_travel_time", "average_travel_time", "stddiv_travel_time",  "shortest_distance", "average_distance_traveled_per_veh", "stddiv_distance_traveled_per_veh"]]
        for o,d in s.od_trips.keys():
            out.append([o.name, d.name, s.od_trips[o,d], s.od_trips_comp[o,d], s.od_tt_free[o,d], s.od_tt_ave[o,d], s.od_tt_std[o,d], s.od_dist_min[o,d], s.od_dist_ave[o,d], s.od_dist_std[o,d]])

        s.df_od = pd.DataFrame(out[1:], columns=out[0])
        return s.df_od
    
    def areas2areas_to_pandas(s, areas, area_names=None):
        """
        Compute the area-wise OD-specific stats and return a pandas DataFrame. It analyzes travel stats between areas (set of nodes).
        
        Parameters
        ----------
        areas : list
            The list of areas. Each area is defined as a list of nodes. The items of area can be Node objects or names of Nodes.
        area_names : list, optional
            The list of names of areas.
            
        Returns
        -------
        pd.DataFrame
        """
        df = s.od_to_pandas()

        o_name_rec = []
        d_name_rec = []
        total_trips_rec = []
        completed_trips_rec = []
        average_travel_time_rec = []
        stddiv_travel_time_rec = []
        average_distance_traveled_rec = []
        stddiv_distance_traveled_rec = []

        average_free_travel_time_rec = []
        average_shortest_distance_rec = []

        areas = [[s.W.get_node(n).name for n in area] for area in areas]
        if area_names == None: 
            area_names = [f"area {i} including {areas[i][0]}" for i in range(len(areas))]

        for i, origs in enumerate(areas):
            for j, dests in enumerate(areas):
                o_name = area_names[i]
                d_name = area_names[j]

                # print(o_name, d_name)

                # group by area: average travel time from origs to dests
                rows = df["orig"].isin(origs) & df["dest"].isin(dests)
                total_tripss = np.array(df["total_trips"][rows])
                average_travel_times = np.array(df["average_travel_time"][rows])
                completed_tripss = np.array(df["completed_trips"][rows])
                var_travel_times = np.array(df["stddiv_travel_time"][rows])**2
                distance_traveleds = np.array(df["average_distance_traveled_per_veh"][rows])
                var_distance_traveleds = np.array(df["stddiv_distance_traveled_per_veh"][rows])**2

                free_travel_time_times = np.array(df["free_travel_time"][rows])
                shortest_distances = np.array(df["shortest_distance"][rows])

                # print(f"{total_tripss = }")
                # print(f"{average_travel_times = }")
                # print(f"{completed_tripss = }")
                # print(f"{var_travel_times = }")
                # print(f"{distance_traveleds = }")
                # print(f"{var_distance_traveleds = }")

                total_trips = total_tripss.sum()
                completed_trips = completed_tripss.sum()

                if total_trips:
                    average_travel_time = (completed_tripss*average_travel_times).sum()/completed_trips
                    var_travel_time = (completed_tripss*var_travel_times).sum()/completed_trips    #wrong! there is a correct formula. TODO: implement
                    stddiv_travel_time = np.sqrt(var_travel_time)

                    average_shortest_distance = (total_tripss*shortest_distances).sum()/total_trips
                else:
                    continue
                    # average_travel_time = np.nan
                    # var_travel_time = np.nan
                    # stddiv_travel_time = np.nan
                    # average_shortest_distance = np.nan

                if completed_trips:
                    average_distance_traveled = (total_tripss*distance_traveleds).sum()/total_trips
                    var_distance_traveled = (total_tripss*distance_traveleds).sum()/total_trips    #wrong!
                    stddiv_distance_traveled = np.sqrt(var_distance_traveled)

                    average_free_travel_time = (completed_tripss*free_travel_time_times).sum()/completed_trips
                else:
                    average_distance_traveled = np.nan
                    var_distance_traveled = np.nan
                    stddiv_distance_traveled = np.nan
                    average_free_travel_time = np.nan

                # print(f"{total_trips = }")
                # print(f"{completed_trips = }")
                # print(f"{average_travel_time = }")
                # print(f"{stddiv_travel_time = }")
                # print(f"{average_distance_traveled = }")
                # print(f"{stddiv_distance_traveled = }")

                o_name_rec.append(o_name)
                d_name_rec.append(d_name)
                total_trips_rec.append(total_trips)
                completed_trips_rec.append(completed_trips)
                average_travel_time_rec.append(average_travel_time)
                stddiv_travel_time_rec.append(stddiv_travel_time)
                average_distance_traveled_rec.append(average_distance_traveled)
                stddiv_distance_traveled_rec.append(stddiv_distance_traveled)
                average_free_travel_time_rec.append(average_free_travel_time)
                average_shortest_distance_rec.append(average_shortest_distance)

        out = [["origin_area", "destination_area", "total_trips", "completed_trips", "average_travel_time", "average_free_travel_time", "average_distance_traveled", "average_shortest_distance"]]
        out += [[o_name_rec[i], d_name_rec[i], total_trips_rec[i], completed_trips_rec[i], average_travel_time_rec[i], average_free_travel_time_rec[i], average_distance_traveled_rec[i], average_shortest_distance_rec[i]] for i in range(len(o_name_rec))]
        
        s.df_areas2areas = pd.DataFrame(out[1:], columns=out[0])
        return s.df_areas2areas

    def area_to_pandas(s, areas, area_names=None, border_include=True):
        """
        Compute traffic stats in area and return as pandas.DataFrame.

        Parameters
        ----------
        areas : list
            The list of areas. Each area is defined as a list of nodes. The items of area can be Node objects or names of Nodes.
        area_names : list, optional
            The list of names of areas.
        border_include : bool, optional
            If set to True, the links on the border of the area are included in the analysis. Default is True.

        Returns
        -------
        pd.DataFrame
        """

        # Precompute DataFrames
        df_links = s.W.analyzer.link_to_pandas()
        df_veh_link = s.W.analyzer.vehicles_to_pandas().drop_duplicates(subset=['name', 'link'])

        # Prepare areas as sets for fast lookup
        areas_set = [{s.W.get_node(n).name for n in area} for area in areas]

        # Initialize result lists
        n_links_rec = []
        traffic_volume_rec = []
        vehicles_remain_rec = []
        total_travel_time_rec = []
        average_delay_rec = []
        average_speed_rec = []
        vehicle_density_rec = []

        # Vectorized approach to process all areas at once
        for area_set in areas_set:
            if border_include:
                rows = df_links["start_node"].isin(area_set) | df_links["end_node"].isin(area_set)
            else:
                rows = df_links["start_node"].isin(area_set) & df_links["end_node"].isin(area_set)

            links = df_links.loc[rows, "link"].unique()

            n_links = links.size
            traffic_volume = df_veh_link[df_veh_link["link"].isin(links)]["name"].nunique() * s.W.DELTAN
            vehicles_remain = df_links.loc[rows, "vehicles_remain"].sum()

            if traffic_volume > 0:
                traffic_volume_rows = (
                            df_links.loc[rows, "traffic_volume"] - df_links.loc[rows, "vehicles_remain"]).values
                total_travel_time = np.sum(df_links.loc[rows, "average_travel_time"].values * traffic_volume_rows)
                total_free_time = np.sum(df_links.loc[rows, "free_travel_time"].values * traffic_volume_rows)
                average_delay = max(total_travel_time / total_free_time - 1, 0)

                # Average speed calculation: total distance / total time
                total_distance = np.sum(df_links.loc[rows, "length"].values * traffic_volume_rows)
                average_speed = total_distance / total_travel_time if total_travel_time > 0 else np.nan

                # Vehicle density calculation: total number of vehicles / total link length
                total_link_length = df_links.loc[rows, "length"].sum()
                vehicle_density = traffic_volume / total_link_length if total_link_length > 0 else np.nan
            else:
                total_travel_time = 0
                total_free_time = 0
                average_delay = np.nan
                average_speed = np.nan
                vehicle_density = np.nan

            # Append the results to lists
            n_links_rec.append(n_links)
            traffic_volume_rec.append(traffic_volume)
            vehicles_remain_rec.append(vehicles_remain)
            total_travel_time_rec.append(total_travel_time)
            average_delay_rec.append(average_delay)
            average_speed_rec.append(average_speed)
            vehicle_density_rec.append(vehicle_density)

        # Create DataFrame from the results
        df_result = pd.DataFrame({
            "area": area_names,
            "n_links": n_links_rec,
            "traffic_volume": traffic_volume_rec,
            "vehicles_remain": vehicles_remain_rec,
            "total_travel_time": total_travel_time_rec,
            "average_delay": average_delay_rec,
            "average_speed": average_speed_rec,
            "vehicle_density": vehicle_density_rec,
        })

        s.df_area = df_result
        return s.df_area

    def vehicle_groups_to_pandas(s, groups, group_names=None):
        """
        Computes the stats of vehicle group and return as a pandas DataFrame.

        Parameters
        ----------
        groups : list
            The list of vehicle groups. Each group is defined as a list of vehicle object.
        group_names : list, optional
            The list of names of vehicle groups.
        
        Returns
        -------
        pd.DataFrame
        """
        df_od = s.W.analyzer.od_to_pandas()

        if group_names == None: 
            group_names = [f"group {i} including {groups[0].name}" for i in range(len(groups))]

        total_trip_rec = []
        completed_trip_rec = []
        average_travel_time_rec = []
        average_delay_rec = []
        std_delay_rec = []
        average_traveled_distance_rec = []
        average_detour_rec = []
        std_detour_rec = []
        averae_speed_rec = []
        std_speed_rec = []
        for i, group in enumerate(groups):
            total_trips = 0
            completed_trips = 0
            travel_times = []
            delays = []
            traveled_distances = []
            detours = []
            speeds = []


            for veh in group:

                total_trips += 1
                if veh.state == "end":
                    completed_trips += 1
                    travel_times.append(veh.travel_time)
                    traveled_distances.append(veh.distance_traveled)
                    
                    free_travel_time = df_od["free_travel_time"][(df_od["orig"]==veh.orig.name) & (df_od["dest"]==veh.dest.name)].values[0]
                    shortest_distance = df_od["shortest_distance"][(df_od["orig"]==veh.orig.name) & (df_od["dest"]==veh.dest.name)].values[0]

                    delays.append(veh.travel_time/free_travel_time)
                    detours.append(veh.distance_traveled/shortest_distance)

                    speeds.append(veh.distance_traveled/veh.travel_time)

                #print(f"{group_names[i]=}, {np.average(travel_times)=}, {np.average(traveled_distances)=}, {np.average(delays)=}, {np.average(detours)=}, {np.std(delays)=}, {np.std(detours)=}, {np.average(speeds)}, {np.std(speeds)}")

            total_trip_rec.append(total_trips)
            completed_trip_rec.append(completed_trips)
            if completed_trips > 0:
                average_travel_time_rec.append(np.average(travel_times))
                average_delay_rec.append(np.average(delays))
                std_delay_rec.append(np.std(delays))
                average_traveled_distance_rec.append(np.average(traveled_distances))
                average_detour_rec.append(np.average(detours))
                std_detour_rec.append(np.std(detours))
                averae_speed_rec.append(np.average(speeds))
                std_speed_rec.append(np.std(speeds))
            else:
                average_travel_time_rec.append(np.nan)
                average_delay_rec.append(np.nan)
                std_delay_rec.append(np.nan)
                average_traveled_distance_rec.append(np.nan)
                average_detour_rec.append(np.nan)
                std_detour_rec.append(np.nan)
                averae_speed_rec.append(np.nan)
                std_speed_rec.append(np.nan)

        df = pd.DataFrame({
            "group": group_names,
            "total_trips": total_trip_rec,
            "completed_trips": completed_trip_rec,
            "average_travel_time": average_travel_time_rec,
            "average_delay_ratio": average_delay_rec,
            "std_delay_ratio": std_delay_rec,
            "average_traveled_distance": average_traveled_distance_rec,
            "average_detour_ratio": average_detour_rec,
            "std_detour_ratio": std_detour_rec,
            "average_speed": averae_speed_rec,
            "std_speed": std_speed_rec,
        })

        s.df_vehicle_groups = df
        
        return s.df_vehicle_groups

    def mfd_to_pandas(s, links=None):
        """
        Compute the MFD-like stats and return as a pandas DataFrame.

        Returns
        -------
        pd.DataFrame
        """
        if links == None:
            links = s.W.LINKS
        s.compute_mfd(links)
        links = [s.W.get_link(link) for link in links]
        links = frozenset(links)

        out = [["t", "network_k", "network_q"]]
        for i in lange(s.W.K_AREA):
            out.append([i*s.W.EULAR_DT, s.W.K_AREA[links][i], s.W.Q_AREA[links][i]])
        s.df_mfd = pd.DataFrame(out[1:], columns=out[0])
        return s.df_mfd

    def link_to_pandas(s):
        """
        Converts the link-level analysis results to a pandas DataFrame.

        Returns
        -------
        pd.DataFrame
        """
        s.link_analysis_coarse()

        out = [["link", "start_node", "end_node", "traffic_volume", "vehicles_remain", "free_travel_time", "average_travel_time", "stddiv_travel_time", "delay_ratio", "length"]]
        for l in s.W.LINKS:
            out.append([l.name, l.start_node.name, l.end_node.name, s.linkc_volume[l], s.linkc_remain[l], s.linkc_tt_free[l], s.linkc_tt_ave[l], s.linkc_tt_std[l], s.linkc_tt_ave[l]/s.linkc_tt_free[l], l.length])
        s.df_linkc = pd.DataFrame(out[1:], columns=out[0])
        return s.df_linkc

    def link_traffic_state_to_pandas(s):
        """
        Compute the traffic states in links and return as a pandas DataFrame.

        Returns
        -------
        pd.DataFrame
        """
        s.compute_edie_state()

        out = [["link", "t", "x", "delta_t", "delta_x", "q", "k", "v"]]
        for l in s.W.LINKS:
            for i in range(l.k_mat.shape[0]):
                for j in range(l.k_mat.shape[1]):
                    out.append([l.name, i*l.edie_dt, j*l.edie_dx, l.edie_dt, l.edie_dx, l.q_mat[i,j], l.k_mat[i,j], l.v_mat[i,j]])
        s.df_link_traffic_state = pd.DataFrame(out[1:], columns=out[0])
        return s.df_link_traffic_state

    def link_cumulative_to_pandas(s):
        """
        Compute the cumulative counts etc. in links and return as a pandas DataFrame.

        Returns
        -------
        pd.DataFrame
        """
        out = [["link", "t", "arrival_count", "departure_count", "actual_travel_time", "instantanious_travel_time"]]
        for link in s.W.LINKS:
            for i in range(s.W.TSIZE):
                out.append([link.name, i*s.W.DELTAT, link.cum_arrival[i], link.cum_departure[i], link.traveltime_actual[i], link.traveltime_instant[i]])
        s.df_link_cumulative = pd.DataFrame(out[1:], columns=out[0])
        return s.df_link_cumulative

    @catch_exceptions_and_warn()
    def output_data(s, fname=None):
        """
        Save all results to CSV files. This is obsolute; not all functions are implemented.
        """
        if fname == None:
            fname = f"out{s.W.name}/data"
        s.basic_to_pandas().to_csv(fname+"_basic.csv", index=False)
        s.od_to_pandas().to_csv(fname+"_od.csv", index=False)
        s.mfd_to_pandas().to_csv(fname+"_mfd.csv", index=False)
        s.link_to_pandas().to_csv(fname+"_link.csv", index=False)
        s.link_traffic_state_to_pandas().to_csv(fname+"_link_traffic_state.csv", index=False)
        s.vehicles_to_pandas().to_csv(fname+"_vehicles.csv", index=False)



from dataclasses import dataclass
import enum, math
from typing import Callable, Dict, Iterable, Tuple, Union
from feasytools import RangeList
from .utils import IntPairList

_INF = float('inf')

@dataclass
class Trip:
    id:str
    depart_time:int
    from_node:str
    to_node:str

    def __repr__(self):
        return str(self)
    
    def __str__(self):
        return f"{self.from_node}->{self.to_node}@{self.depart_time}"
    

class VehStatus(enum.IntEnum):
    """
    Vehicle status enumeration
        Driving: Driving to destination or charging station
        Pending: Already notified SUMO to start, but the vehicle has not yet started in SUMO
        Charging: Charging at charging station
        Parking: Parking (between two trips, or before the trip, or after the trip)
        Depleted: Battery depleted
    """
    Driving = 0
    Pending = 1
    Charging = 2
    Parking = 3
    Depleted = 4

def _EqualChargeRate(rate: float, ev: 'EV') -> float:
    return rate

def _LinearChargeRate(rate: float, ev: 'EV') -> float:
    if ev.SOC <= 0.8:
        return rate
    return rate * (3.4 - 3 * ev.SOC)

class ChargeRatePool:
    """Charging rate correction function pool"""
    _pool:'Dict[str, Callable[[float, EV], float]]' = {
        "Equal":_EqualChargeRate, 
        "Linear":_LinearChargeRate,
    }

    @staticmethod
    def add(name: str, func: 'Callable[[float, EV], float]'):
        """Add charging rate correction function"""
        ChargeRatePool._pool[name] = func

    @staticmethod
    def get(name: str) -> 'Callable[[float, EV], float]':
        """Get charging rate correction function"""
        return ChargeRatePool._pool[name]

class EV:
    """Electric Vehicle Class"""
    def __init__(
        self,
        id: str,
        trips: Iterable[Trip],
        eta_c: float,
        eta_d: float,
        bcap: float,
        soc: float,
        c: float,
        rf: float,
        rs: float,
        rv: float,
        omega: float,
        kr: float,
        kf: float,
        ks: float,
        kv: float,
        rmod: str = "Linear",
        sc_time: Union[None, IntPairList, RangeList] = None,
        max_sc_cost: float = 100.0,
        v2g_time: Union[None, IntPairList, RangeList] = None,
        min_v2g_earn: float = 0.0,
        cache_route: bool = False,
    ):
        self._id = id                   # Vehicle ID
        self._sta = VehStatus.Parking   # Vehicle status
        self._cs = None                 # Target charging station
        self._cost = 0                  # Total charging cost of the vehicle, $
        self._earn = 0                  # Total discharge revenue of the vehicle, $

        self._bcap = bcap               # Battery capacity, kWh
        assert 0.0 <= soc <= 1.0
        self._elec = soc * bcap         # Current battery capacity, kWh
        self._consumption = c / 1000    # Energy consumption per unit distance, kWh/m
        self._efc_rate = rf / 3600      # Expected fast charge power, kWh/s
        self._esc_rate = rs / 3600      # Expected slow charge power, kWh/s
        self._v2g_rate = rv / 3600      # Maximum reverse power, kWh/s
        self._eta_charge = eta_c        # Charging efficiency
        self._eta_discharge = eta_d     # Discharge efficiency
        self._chrate_mod = ChargeRatePool.get(rmod)
                                        # Charging rate correction function
        self._sc_time = sc_time if isinstance(sc_time, RangeList) else RangeList(sc_time)
                                        # RangeList of slow charging time, None means all day
        self._max_sc_cost = max_sc_cost # Maximum slow charging cost, $/kWh
        self._v2g_time = v2g_time if isinstance(v2g_time, RangeList) else RangeList(v2g_time)
                                        # RangeList of V2G time, None means all day
        self._min_v2g_earn = min_v2g_earn
                                        # Minimum V2G cost, $/kWh

        self._rate = 0                  # Actual charging power, kWh/s
        self._chtar = bcap              # When fast charging, how much kWh to charge before leaving

        self._dis = 0                   # Distance traveled, m
        self._trips = tuple(trips)      # Vehicle trip list
        self._trip_index = 0            # Current trip number (index)

        self._w = omega                 # Decision parameter
        assert 1 <= kr <= 2
        self._krel = kr                 # Tolerance coefficient
        assert 0 < kf < 1
        self._kfc = kf                  # User selects SoC for fast charging
        assert kf <= ks < 1
        self._ksc = ks                  # User selects SoC for slow charging
        assert ks < kv
        self._kv2g = kv                 # SoC where the user is willing to join V2G

        self._cache_route = cache_route # Whether to cache the route

        self.__tmp_pc_max = _INF # Temporary variable, maximum charging power kWh/s
        self.__tmp_pd = self._v2g_rate # Temporary variable, maximum discharging power kWh/s
    
    def set_temp_max_pc(self, pc: float):
        """Set temporary maximum charging power kWh/s. 
        This function must be called in MaxPCAllocator."""
        self.__tmp_pc_max = pc

    def set_temp_pd(self, pd: float):
        """Set temporary discharging power kWh/s.
        This function must be called in V2GAllocator."""
        self.__tmp_pd = pd

    @property
    def estimated_charge_time(self) -> float:
        """
        Time required to complete charging at the current charge level, target charge level and charging rate
        """
        if self._rate > 0:
            return max((self._chtar - self._elec) / self._rate, 0)
        else:
            return math.inf

    @property
    def elec(self) -> float:
        """Current battery capacity kWh"""
        return self._elec

    @property
    def SOC(self) -> float:
        """SoC of the battery (percentage)"""
        return self._elec / self._bcap

    @property
    def omega(self) -> float:
        """Decision parameter for selecting fast charging station"""
        return self._w

    @omega.setter
    def omega(self, val: float):
        self._w = val

    @property
    def krel(self) -> float:
        """Tolerance coefficient"""
        return self._krel

    @krel.setter
    def krel(self, val: float):
        assert 1.0 <= val <= 2.0
        self._krel = val

    @property
    def kfc(self) -> float:
        """Select the SOC threshold for fast charging"""
        return self._kfc

    @kfc.setter
    def kfc(self, val: float):
        assert 0.0 <= val <= self._ksc
        self._kfc = val

    @property
    def ksc(self) -> float:
        """Select the SOC threshold for slow charging"""
        return self._ksc

    @ksc.setter
    def ksc(self, val: float):
        assert self._kfc <= val <= 1.0
        self._ksc = val

    @property
    def kv2g(self) -> float:
        """Select the SOC threshold for slow charging"""
        return self._kv2g

    @kv2g.setter
    def kv2g(self, val: float):
        assert self._ksc < val
        self._kv2g = val

    @property
    def charge_target(self) -> float:
        """
        Charging target, that is, charge to this level and leave the fast charging station (kWh)
        """
        return self._chtar

    @charge_target.setter
    def charge_target(self, val):
        assert 0 <= val <= self._bcap
        self._chtar = val

    @property
    def eta_charge(self) -> float:
        """Vehicle charging efficiency"""
        return self._eta_charge

    @property
    def eta_discharge(self) -> float:
        """Vehicle discharge efficiency"""
        return self._eta_discharge

    @property
    def pc_actual(self) -> float:
        """Vehicle's actual charging rate, kWh/s"""
        return self._rate

    rate = pc_actual

    @property
    def max_v2g_rate(self) -> float:
        """Vehicle's maximum V2G reverse power rate, kWh/s"""
        return self._v2g_rate
    
    max_pd = max_v2g_rate

    def stop_charging(self):
        """Stop charging: set charging rate to 0"""
        self._rate = 0

    @property
    def status(self) -> VehStatus:
        """Current vehicle status"""
        return self._sta

    @status.setter
    def status(self, val):
        self._sta = val

    @property
    def target_CS(self) -> Union[None, str]:
        """
        Name of the target fast charging station. When this item is None, it means that it isn't guided to a CS
        """
        return self._cs

    @target_CS.setter
    def target_CS(self, val):
        self._cs = val

    @property
    def ID(self) -> str:
        """Vehicle's string ID"""
        return self._id

    @property
    def full_battery(self) -> float:
        """Vehicle's battery capacity kWh"""
        return self._bcap

    @property
    def battery(self) -> float:
        """Vehicle's battery level kWh"""
        return self._elec

    @property
    def consumption(self) -> float:
        """Vehicle's energy consumption kWh/m"""
        return self._consumption

    @property
    def odometer(self) -> float:
        """
        Distance traveled by the vehicle in this trip (m), note that leaving the charging station is considered a new trip
        """
        return self._dis

    @property
    def minimum_v2g_earn(self) -> float:
        """The minimum V2G earn user willing to join V2G, $/kWh"""
        return self._min_v2g_earn
    
    @property
    def maximum_slow_charge_cost(self) -> float:
        """The maximum slow charging cost willing to join slow charge, $/kWh"""
        return self._max_sc_cost
    
    @property
    def v2g_time(self) -> RangeList:
        """The time range that the user is willing to join V2G. None means all day"""
        return self._v2g_time
    
    @property
    def slow_charge_time(self) -> RangeList:
        """The time range that the user is willing to join slow charge. None means all day"""
        return self._sc_time
    
    def clear_odometer(self):
        """Before the trip starts, clear the odometer"""
        self._dis = 0

    def drive(self, new_dis: float):
        """
        Update the battery SOC and odometer in driving state
        """
        # Since SUMO save and load may lead to error, so -1.0 to tolerate the error
        assert new_dis >= self._dis - 1.0, f"EV {self._id}: self._dis = {self._dis:.8f} > new_dis = {new_dis:.8f}"
        self._elec -= (new_dis - self._dis) * self._consumption
        self._dis = new_dis

    def charge(self, sec: int, unit_cost: float, pc:float) -> float:
        """
        Charge the battery for sec seconds at given charging power pc kWh/s, 
        and return the actual charging amount (kWh) (considering losses)
        After charging, temporary maximum charging power is reset to infinity
        """
        _elec = self._elec
        self._rate = min(self._chrate_mod(pc, self), self.__tmp_pc_max)
        self.__tmp_pc_max = _INF
        self._elec += self._rate * sec * self._eta_charge
        if self._elec > self._bcap:
            self._elec = self._bcap
        if self._elec < 0:
            self._elec = 0
        delta_elec = self._elec - _elec
        self._cost += (delta_elec / self._eta_charge) * unit_cost
        return delta_elec

    def discharge(self, sec: int, unit_earn: float) -> float:
        """
        Discharge the battery for sec seconds, whose discharging power is set by set_temp_pd(),
        and return the actual discharge amount kWh (considering losses)
        After discharging, temporary discharging power is reset to the maximum discharging power
        """
        _elec = self._elec
        self._elec -= self.__tmp_pd * sec
        self.__tmp_pd = 0
        if self.SOC <= self._kv2g:
            self._elec = self._bcap * self._kv2g
        delta_elec = (_elec - self._elec) * self._eta_discharge
        assert delta_elec >= 0
        self._earn += delta_elec * unit_earn
        return delta_elec

    def willing_to_v2g(self, t:int, e:float) -> bool:
        """
        User determines whether the vehicle is willing to v2g
            t: current time
            e: current V2G earn, $/kWh
        """
        return self.SOC > self._kv2g and e >= self._min_v2g_earn and (self._v2g_time.__contains__(t) if self._v2g_time else True)
    
    def willing_to_slow_charge(self, t:int, c:float) -> bool:
        """
        User determines whether the vehicle is willing to slow charge
            t: current time
            c: current slow charge cost, $/kWh
        """
        return c <= self._max_sc_cost and self._sc_time.__contains__(t)
    
    @property
    def trips(self) -> Tuple[Trip, ...]:
        '''Get the list of trips for the vehicle'''
        return self._trips

    @property
    def trips_count(self) -> int:
        '''Get the number of trips for the vehicle'''
        return len(self._trips)
    
    @property
    def trip(self) -> Trip:
        '''Get the current trip'''
        return self._trips[self._trip_index]

    @property
    def trip_id(self) -> int:
        '''Get the ID of the current trip (indexed from 0)'''
        return self._trip_index

    def next_trip(self) -> int:
        """
        Increment the trip ID by 1 and return the trip ID. If it is already the last trip, return -1.
        """
        if self._trip_index == len(self.trips) - 1:
            return -1
        self._trip_index += 1
        return self._trip_index

    @property
    def max_mileage(self) -> float:
        """
        Maximum mileage under the current battery level (m)
        """
        return self._elec / self._consumption

    def is_batt_enough(self, dist: float) -> bool:
        """
        User determines whether the current battery level is sufficient to travel a distance of dist
        """
        return self.max_mileage >= self._krel * dist

    def brief(self):
        """Get a brief description of this vehicle"""
        return f"{self._id},{self.SOC*100:.1f}%,{self._elec:.1f}kWh,{self._trip_index}"

    def __repr__(self):
        return f"EV[ID='{self._id}', Status={self._sta}, Dist={self._dis}m, SOC={self.SOC*100:.1f}%, Bcap={self._bcap}kWh, ChTar={self._chtar}kWh, TarCS={self._cs}, Consump={self._consumption}KWh/m]"

    def __str__(self):
        return repr(self)

__all__ = ["Trip", "VehStatus", "EV", "ChargeRatePool"]


import os
import sys
from typing import Dict, Iterable, Tuple, Type, Union, Optional, TypeVar, Generic, List
from feasytools import RangeList
from sklearn.neighbors import KDTree
from concurrent.futures import ThreadPoolExecutor, as_completed
from ..locale import Lang
from .ev import EV
from .utils import ReadXML
from .params import *
from .cs import SCS, FCS

T_CS = TypeVar("T_CS", FCS, SCS)
CS_Type = Type[T_CS]


def LoadCSList(filePath:str, csType:CS_Type) -> List[T_CS]:
    assert csType == FCS or csType == SCS
    _cs = []
    root = ReadXML(filePath).getroot()
    if root is None: raise ValueError("Invalid CS file")
    for cs_node in root:
        par_pbuy = ()
        par_psell = ()
        par_off = []
        for cfg in cs_node:
            if cfg.tag == "pbuy" or cfg.tag == "psell":
                tl = [int(itm.attrib["btime"]) for itm in cfg]
                pl = [float(itm.attrib["price"]) for itm in cfg]
                if cfg.tag == "pbuy":
                    par_pbuy = (tl, pl)
                else:
                    par_psell = (tl, pl)
            elif cfg.tag == "offline":
                par_off = RangeList(cfg)
            else:
                raise ValueError(Lang.CSLIST_INVALID_TAG.format(cfg.tag))
        if par_pbuy is None:
            raise ValueError(Lang.CSLIST_PBUY_NOT_SPECIFIED)
        _cs.append(
            csType(
                name = cs_node.attrib["node"],
                slots = int(cs_node.attrib["slots"]),
                bus = cs_node.attrib.get("bus", "None"),
                max_pc = float(cs_node.attrib.get("max_pc", "inf")) / 3600,
                max_pd = float(cs_node.attrib.get("max_pd", "inf")) / 3600,
                x = float(cs_node.attrib.get("x", "inf")),
                y = float(cs_node.attrib.get("y", "inf")),
                offline = par_off, 
                price_buy = par_pbuy, 
                price_sell = par_psell,
                pc_alloc = cs_node.attrib.get("pcalloc", "Average"),
                pd_alloc = cs_node.attrib.get("pdalloc", "Average"),
            )
        )
    return _cs

_LoadCSList = LoadCSList

class CSList(Generic[T_CS]):
    """CS List. Index starts from 0."""
    _cs: List[T_CS]

    def __iter__(self):
        return self._cs.__iter__()

    def __init__(
        self,
        *,
        csList: Optional[List[T_CS]] = None,
        filePath: Optional[str] = None,
        csType: CS_Type = FCS,
    ):
        """
        Initialize
            csList: CS list
            filePath: file name of CS
            csType: CS type, must be FCS or SCS
        Provide csList, or provide filePath and csType together
        """
        if csList is not None:
            flagF = False
            flagS = False
            for cs in csList:
                if isinstance(cs, FCS):
                    flagF = True
                elif isinstance(cs, SCS):
                    flagS = True
                else:
                    raise TypeError(Lang.CSLIST_INVALID_ELEMENT)
            if flagF and flagS and not ALLOW_MIXED_CSTYPE_IN_CSLIST:
                raise TypeError(Lang.CSLIST_MIXED_ELEMENT)
            self._cs = csList
        elif filePath is not None and csType is not None:
            self._cs = _LoadCSList(filePath, csType)
        else:
            raise TypeError(Lang.CSLIST_INVALID_INIT_PARAM)
        self._n = len(self._cs)
        self._remap: Dict[str, int] = {}  # Number of a CS with a name
        self._veh: Dict[str, int] = {}  # Number of the charging station to which a vehicle belongs
        self._cs_names = [cs.name for cs in self._cs]
        self._cs_slots = [cs.slots for cs in self._cs]
        self.__v2g_cap_res: List[float] = [0.0] * self._n
        self.__v2g_cap_res_time:int = -1
        self.__v2g_demand: List[float] = []
        for i, cs in enumerate(self._cs):
            self._remap[cs.name] = i
        self.create_kdtree()
        self.create_pool()

    def shutdown_pool(self):
        """Shut down the thread pool"""
        if self.__pool:
            self.__pool.shutdown(wait=True)
            self.__pool = None
    
    def create_pool(self):
        """Create the thread pool"""
        cpu_count = os.cpu_count() or 1
        if hasattr(sys, "_is_gil_enabled") and not sys._is_gil_enabled() and cpu_count > 1:
            self.__pool = ThreadPoolExecutor(cpu_count)
        else:
            self.__pool = None
    
    def create_kdtree(self):
        pts = []
        for cs in self._cs:
            if cs._x == float("inf") or cs._y == float("inf"):
                self._kdtree = None
                break
            pts.append((cs._x, cs._y))
        else:
            self._kdtree = KDTree(pts) if pts else None
    
    def select_near(self, pos: Tuple[float, float], n: int = 1) -> Iterable[int]:
        """
        Select the n nearest charging station numbers to (x, y).
            pos: coordinate
            n: Number of selections
        Return
            Selected indices. If n is greater than or equal to the total number of charging stations, all charging stations are returned.
        """
        if n >= self._n or self._kdtree is None:
            return range(self._n)
        dist, idx = self._kdtree.query(pos, k=n, return_distance=True)
        return idx.reshape(-1)

    def index(self, name: str) -> int:
        """
        Get the index of the CS by its name. If the name does not exist, KeyError will be raised.
        """
        return self._remap[name]

    def get_CS_names(self) -> List[str]:
        """List of names of all charging stations"""
        return self._cs_names

    def get_online_CS_names(self, t: int) -> List[str]:
        """
        List of names of all available charging stations at t seconds
        """
        return [cs.name for cs in self._cs if cs.is_online(t)]

    def get_prices_at(self, t: int) -> List[float]:
        """
        List of prices of all charging stations at t seconds
        """
        return [cs.pbuy(t) for cs in self._cs]

    def get_online_prices_at(self, t: int) -> List[float]:
        """
        List of prices of all available charging stations at t seconds
        """
        return [cs.pbuy(t) for cs in self._cs if cs.is_online(t)]

    def get_veh_count(self, only_charging=False) -> List[int]:
        """
        List of the number of vehicles at all charging stations. When only_charging is True, only the number of vehicles being charged is returned.
        """
        return [cs.veh_count(only_charging) for cs in self._cs]

    def get_online_veh_count(self, t: int, only_charging=False) -> List[int]:
        """
        List of the number of vehicles at all available charging stations at t seconds. When only_charging is True, only the number of vehicles being charged is returned.
        """
        return [cs.veh_count(only_charging) for cs in self._cs if cs.is_online(t)]

    def get_slots_of(self) -> List[int]:
        """
        List of the number of charging piles at all charging stations
        """
        return self._cs_slots

    def get_online_slots_of(self, t: int) -> List[int]:
        """
        List of the number of charging piles at all available charging stations at t seconds
        """
        return [cs.slots for cs in self._cs if cs.is_online(t)]

    def get_Pd(self, k: float = 1) -> List[float]:
        """
        List of charging power of all charging stations, default unit kWh/s
            k: Unit conversion coefficient, default is 1, indicating no conversion;
               3600 indicates conversion to kW; 3600/Sb_kVA indicates conversion to per unit value
        """
        return [cs.Pd * k for cs in self._cs]

    def get_online_Pd(self, t: int, k: float = 1) -> List[float]:
        """
        List of charging power of all available charging stations
            t: Determination time
            k: Unit conversion coefficient, default is 1, indicating no conversion;
               3600 indicates conversion to kW; 3600/Sb_kVA indicates conversion to per unit value
        """
        return [cs.Pd * k for cs in self._cs if cs.is_online(t)]

    def get_Pc(self, k: float = 1) -> List[float]:
        """
        List of discharge power of all charging stations
            k: Unit conversion coefficient, default is 1, indicating no conversion;
            3600 indicates conversion to kW; 3600/Sb_kVA indicates conversion to per unit value
        """
        return [cs.Pc * k for cs in self._cs]

    def get_online_Pc(self, t: int, k: float = 1) -> List[float]:
        """
        List of discharge power of all available charging stations
            t: Determination time
            sb: Unit conversion coefficient, default is 1, indicating no conversion; 
               3600 indicates conversion to kW; 3600/Sb_kVA indicates conversion to per unit value
        """
        return [cs.Pc * k for cs in self._cs if cs.is_online(t)]

    def add_veh(self, veh:EV, cs: Union[int, str]) -> bool:
        """
        Add vehicles to the specified charging station
            veh_id: Vehicle ID
            cs: Charging station name or string
        Return
            True if added successfully, False if the vehicle is already charging, KeyError if the charging station does not exist.
        """
        if not isinstance(cs, int):
            cs = self.index(cs)
        ret = self._cs[cs].add_veh(veh)
        if ret:
            self._veh[veh.ID] = cs
        return ret

    def pop_veh(self, veh: EV) -> bool:
        """
        Remove vehicles from the charging station
            veh: Vehicle object
        Return:
            True if removed successfully, False if the vehicle does not exist.
        """
        try:
            cs = self._veh[veh.ID]
        except KeyError:
            return False
        del self._veh[veh.ID]
        self._cs[cs].pop_veh(veh)
        return True

    def has_veh(self, veh_id: str) -> bool:
        """
        Check if there is a vehicle with the specified ID
            veh_id: Vehicle ID
        Return
            True if exists, False if not.
        """
        return veh_id in self._veh

    def CS_index(self, cs_name: str) -> int:
        """
        Check the index (starting from 0) of the charging station with the specified name
            cs_name: Charging station name
        Return
            Index. If it does not exist, return -1.
        """
        try:
            return self._cs_names.index(cs_name)
        except ValueError:
            return -1

    def is_charging(self, veh: EV) -> bool:
        """
        Get the charging status of the vehicle. If the vehicle does not exist, a ValueError will be raised.
            veh: Vehicle object
        Return
            True if charging, False if waiting.
        """
        cs = self._veh[veh.ID]
        return self._cs[cs].is_charging(veh)

    def __getitem__(self, indices: Union[str, int]) -> T_CS:
        """
        Get the index of given charging station. It can be either indices or a charging station name.
        """
        if isinstance(indices, str):
            indices = self._remap[indices]
        return self._cs[indices]

    def get_V2G_cap(self, t: int) -> List[float]:
        """
        Get the maximum V2G return power (considering losses) of each charging station at the current moment, in kWh/s
            t: Indicates the current moment
        """
        if t == self.__v2g_cap_res_time:
            return self.__v2g_cap_res
        self.__v2g_cap_res = [cs.get_V2G_cap(t) for cs in self._cs]
        self.__v2g_cap_res_time = t
        return self.__v2g_cap_res

    def set_V2G_demand(self, v2g_demand: List[float]):
        """Set V2G demand"""
        assert len(v2g_demand) == len(self._cs) or len(v2g_demand) == 0
        self.__v2g_demand = v2g_demand

    def __update_cs(self, l:int, r:int, sec:int, cur_time:int, v2g_demand:List[float]) -> List[EV]:
        lst = []
        for i in range(l, r):
            lst.extend(self._cs[i].update(sec, cur_time, v2g_demand[i]))
        return lst
    
    def update(self, sec: int, cur_time: int):
        """
        Charge and V2G discharge the EV with the current parameters.
            sec: Charging duration is sec seconds
            cur_time: Current time (seconds)
        Return
            Vehicle list that has completed charging (power is calculated in kWh/s)
        """
        if len(self.__v2g_demand) > 0:  # Has V2G demand
            assert len(self.__v2g_demand) == len(self._cs)
            v2g_demand = self.__v2g_demand
        else:
            v2g_demand = [0.0] * len(self._cs)
        
        ret:List[EV] = []

        if self.__pool:
            # Use multithreading to speed up
            vcnt = 0; l = 0; r = 0
            futures = []
            for i, cs in enumerate(self._cs):
                vcnt += len(cs)
                if vcnt >= 50:  # Each thread processes at least 50 vehicles
                    r = i + 1
                    futures.append(self.__pool.submit(self.__update_cs, l, r, sec, cur_time, v2g_demand))
                    l = r
                    vcnt = 0
            if l < len(self._cs):
                futures.append(self.__pool.submit(self.__update_cs, l, len(self._cs), sec, cur_time, v2g_demand))

            # Collect results
            for f in as_completed(futures):
                lst = f.result()
                for ev in lst:
                    del self._veh[ev.ID]
                ret.extend(lst)
            
        else:
            # Single thread
            for cs, pd in zip(self._cs, v2g_demand):
                lst = cs.update(sec, cur_time, pd)
                for ev in lst:
                    del self._veh[ev.ID]
                ret.extend(lst)
        return ret

    def __len__(self):
        return len(self._cs)

    def __repr__(self):
        return f"CSList[{','.join(map(str,self._cs))}]"
    
    def __str__(self):
        return repr(self)

__all__ = [
    "LoadCSList", "CSList"
]


from collections import defaultdict
from itertools import chain
from feasytools import LangLib
from fpowerkit import IslandResult, DistFlowSolver, Estimator
from ..traffic.cs import CS
from .pdn import PluginPDN
from .base import *

_locale = LangLib(["zh_CN","en"])
_locale.SetLangLib("zh_CN",
    DESCRIPTION = "过流保护",
    ERROR_NO_PDN = "过流保护依赖于PDN插件",
    ERROR_SMART_CHARGE = "启用有序充电时过流保护不可用",
)
_locale.SetLangLib("en",
    DESCRIPTION = "Over-current protection",
    ERROR_NO_PDN = "Over-current protection depends on PDN plugin",
    ERROR_SMART_CHARGE = "Over-current protection is not available when smart charging is enabled",
)

class PluginOvercurrent(PluginBase[None]):
    @property
    def Description(self)->str:
        return _locale["DESCRIPTION"]
    
    def _save_state(self) -> object:
        '''Save the plugin state'''
        return None
    
    def _load_state(self,state:object) -> None:
        '''Load the plugin state'''

    @staticmethod
    def ElemShouldHave() -> ConfigDict:
        '''Get the plugin configuration item list'''
        return ConfigDict()
    
    def Init(self, elem:ET.Element, inst:TrafficInst, work_dir:Path, res_dir:Path, plg_deps:'List[PluginBase]')->None:
        self.__file = open(str(res_dir / "current_protect.log"), "w")
        self.SetPreStep(self._work)
        self.SetPostSimulation(self.__file.close)
        assert len(plg_deps) == 1 and isinstance(plg_deps[0], PluginPDN), _locale["ERROR_NO_PDN"]
        self.__pdn = plg_deps[0]
        if self.__pdn.isSmartChargeEnabled():
            raise RuntimeError(_locale["ERROR_SMART_CHARGE"])
        if self.__pdn.Solver.est != Estimator.DistFlow:
            raise RuntimeError("Over-current protection only works with DistFlowSolver")
        self.__csatb:Dict[str, List[CS]] = defaultdict(list)
        for cs in chain(inst.SCSList, inst.FCSList):
            self.__csatb[cs._bus].append(cs)
        self.__csatb_closed:Dict[str, bool] = {
            b:False for b in self.__csatb
        }
    
    def _close(self, b:str):
        '''
        Force shutdown all charging stations at bus b
        '''
        if self.__csatb_closed[b]: return
        for cs in self.__csatb[b]:
            cs.force_shutdown()
        print(f"CS {','.join(map(lambda x:x.name, self.__csatb[b]))} forced shutdown at bus {b}.", file=self.__file)

    def _work(self,_t:int,/,sta:PluginStatus)->Tuple[bool, None]:
        '''
        Get the V2G demand power of all bus with slow charging stations at time _t, unit kWh/s, 3.6MW=3600kW=1kWh/s
        '''
        if sta == PluginStatus.EXECUTE:
            p = self.__pdn
            if p.LastPreStepSucceed:
                svr = p.Solver
                assert isinstance(svr.est, DistFlowSolver), "Over-current protection only works with DistFlowSolver"
                if len(svr.est.OverflowLines) > 0:
                    print(f"[{_t}] Overcurrent protection triggered: ", svr.est.OverflowLines, file=self.__file)
                    for l in svr.est.OverflowLines:
                        ln = p.Grid.Line(l)
                        ln.P = 0.
                        ln.Q = 0.
                        ln.I = 0.
                    svr.est.UpdateGrid(p.Grid, cut_overflow_lines=True)
                    self.__island_closed = [False] * len(svr.est.Islands)
                    svr.solve(_t)
                for i, il in enumerate(svr.est.Islands):
                    if il.result == IslandResult.Failed and not self.__island_closed[i]:
                        self.__island_closed[i] = True
                        for b in il.Buses: self._close(b)
                ret = True, None
            else:
                ret = False, None
        elif sta == PluginStatus.OFFLINE:
            ret = True, None
        elif sta == PluginStatus.HOLD:
            ret = True, self.LastPreStepResult
        return ret


from v2simux_gui.common import *
from ..mainbox.controls import ScrollableTreeView
from .utils import *


class ParamsEditor(Toplevel):
    def __init__(self, data:Dict[str,str]):
        super().__init__()
        self.data = data
        self.tree = ScrollableTreeView(self, allowSave=False)
        self.tree['show'] = 'headings'
        self.tree["columns"] = ("lb", "rb")
        self.tree.column("lb", width=120, stretch=NO)
        self.tree.column("rb", width=120, stretch=NO)
        self.tree.heading("lb", text=_L["PARAM_NAME"])
        self.tree.heading("rb", text=_L["PARAM_VALUE"])
        self.tree.pack(fill="both", expand=True)
        for l,r in data.items():
            self.tree.insert("", "end", values=(l, r))
        self.tree.setColEditMode("lb", EditMode.combo([
            'b','e','l','no-plg','seed','gen-veh','gen-fcs','gen-scs','plot'
        ]))
        self.tree.setColEditMode("rb", EditMode.entry())
        self.fr = Frame(self)
        self.fr.pack(fill="x", expand=False)
        self.btn_add = Button(self.fr, text=_L["ADD"], command=self.add, width=6)
        self.btn_add.grid(row=0,column=0,pady=3,sticky="w")
        self.btn_del = Button(self.fr, text=_L["DELETE"], command=self.delete, width=6)
        self.btn_del.grid(row=0,column=1,pady=3,sticky="w")
        self.btn_moveup = Button(self.fr, text=_L["CLEAR"], command=self.tree.clear, width=6)
        self.btn_moveup.grid(row=0,column=2,pady=3,sticky="w")
        self.btn_save = Button(self.fr, text=_L["SAVE_AND_CLOSE"], command=self.save)
        self.btn_save.grid(row=0,column=3,padx=3,pady=3,sticky="e")
    
    def add(self):
        self.tree.insert("", "end", values=("no-plg",""))
    
    def delete(self):
        for i in self.tree.tree.selection():
            self.tree.delete(i)
    
    def save(self):
        self.data = self.getAllData()
        self.destroy()

    def getAllData(self) -> Dict[str,str]:
        res:Dict[str,str] = {}
        for i in self.tree.get_children():
            x = self.tree.tree.item(i, "values")
            res[x[0]] = x[1]
        return res

__all__ = ["ParamsEditor"]


import os
from collections import defaultdict
from itertools import chain
from warnings import warn
from feasytools import TimeImplictFunc
from fpowerkit import Grid, GridSolveResult, CombinedSolver, Estimator, Calculator, DistFlowSolver
from feasytools import LangLib
from ..traffic import DetectFiles, CS
from .base import *

_locale = LangLib(["zh_CN","en"])
_locale.SetLangLib("zh_CN",
    DESCRIPTION = "配电网模型",
    ERROR_NO_GRID = "未指定电网文件",
    ERROR_CS_NODE_NOT_EXIST = "错误: 充电站{0}的母线{1}在电网中不存在",
    ERROR_SOLVE_FAILED = "求解失败",
    ERROR_LAST_FAILED = "上次求解失败",
    PDN_SOLVE_FAILED = "[PDN]在时间{0}未能求解配电网. 总共失败{1}次",
    PDN_SOLVE_FAILED_CNT = "[PDN]总共求解失败{}次",
    PDN_SOLVE_OK_SINCE = "[PDN]从时间{0}开始成功求解配电网. 之前总共失败{1}次"
)
_locale.SetLangLib("en",
    DESCRIPTION = "Power distribution network model",
    ERROR_NO_GRID = "No grid file specified",
    ERROR_CS_NODE_NOT_EXIST = "Error: The bus {1} of charging station {0} does not exist in the grid",
    ERROR_SOLVE_FAILED = "Failed to solve",
    ERROR_LAST_FAILED = "Last solve failed",
    PDN_SOLVE_FAILED = "[PDN] Failed to solve the PDN at time: {0}. Totally failed for {1} times",
    PDN_SOLVE_FAILED_CNT = "[PDN] Failed to solve the PDN for {} times",
    PDN_SOLVE_OK_SINCE = "[PDN] Succeed to solve the PDN since time: {0}. Previously failed for {1} times"
)

def _sv(x)->float:
    assert x is not None, _locale["ERROR_SOLVE_FAILED"]
    return x

class PluginPDN(PluginBase[float], IGridPlugin):
    @property
    def Description(self)->str:
        return _locale["DESCRIPTION"]
    
    @staticmethod
    def __create_closure(cslist: List[CS], Sb_MVA: float):
        def _sumP():
            return sum(c.Pc_MW for c in cslist)/Sb_MVA
        return _sumP
    
    def _save_state(self) -> object:
        '''Save the plugin state'''
        return None
        
    def _load_state(self, state:object) -> None:
        '''Load the plugin state'''

    @staticmethod
    def ElemShouldHave() -> ConfigDict:
        '''Get the plugin configuration item list'''
        EST = [Estimator.DistFlow.value, Estimator.LinDistFlow.value, Estimator.LinDistFlow2.value]
        CAL = [Calculator.OpenDSS.value, Calculator.Newton.value, Calculator.NoneSolver.value]
        return ConfigDict([
            ConfigItem("estimator", EditMode.COMBO, f"Estimator of the solver, must be {EST}", Estimator.DistFlow.value, combo_values=EST),
            ConfigItem("calculator", EditMode.COMBO, f"Calculator of the solver, can be {CAL}", Calculator.NoneSolver.value, combo_values=CAL),
            ConfigItem("MLRP", EditMode.ENTRY, "Maximum load reduction percentage", 0.5),
            ConfigItem("source_bus", EditMode.ENTRY, "Source bus of the grid", ""),
            ConfigItem("DecBuses", EditMode.ENTRY, "List of buses for load reduction, separated by commas", ""),
            ConfigItem("SmartCharge", EditMode.COMBO, "Whether to enable smart charging, 'YES' or 'NO'", "NO", combo_values=['YES', 'NO']),
        ])
    
    def Init(self, elem:ET.Element, inst:TrafficInst, work_dir:Path,
            res_dir:Path, plg_deps:'List[PluginBase]') -> float:
        '''Initialize the plugin from the XML element'''
        self.__inst = inst
        self.SetPreStep(self.PreStep)
        self.__fh = open(res_dir / "pdn_res.log", "w")
        self.SetPostSimulation(self.__fh.close)

        res = DetectFiles(str(work_dir))
        assert res.grid, _locale["ERROR_NO_GRID"]
        self.__gr = Grid.fromFile(res.grid, True)
        decs = list(map(lambda x:x.strip(), elem.get("DecBuses","").split(",")))
        if elem.get("SmartCharge", "NO") == "NO":
            decs.clear()
        est = elem.get("estimator", "DistFlow")
        if est != Estimator.DistFlow.value and len(decs)>0:
            warn("Load reduction only supported in DistFlowSolver. Ignoring DecBuses.", UserWarning)
            decs.clear()
        cal = Calculator(elem.get("calculator", "None"))
        if cal == Calculator.OpenDSS:
            source_bus = elem.get("source_bus", "")
        else:
            source_bus = ""
        self.__save_to = str(res_dir / "pdn_logs")
        self.__sol = CombinedSolver(self.__gr,
            estimator=Estimator(est),
            calculator=Calculator(cal),
            mlrp=float(elem.get("MLRP","0.5")),
            source_bus=source_bus,
            default_saveto=self.__save_to
        )
        self.__sol.SetErrorSaveTo(self.__save_to)
        self.__badcnt = 0
        self.__pds:dict[str, List[CS]] = defaultdict(list)
        for c in chain(inst.FCSList, inst.SCSList):
            if not c.bus in self.__gr.BusNames:
                raise ValueError(_locale["ERROR_CS_NODE_NOT_EXIST"].format(c.name,c.bus))
            self.__pds[c.bus].append(c)
        for b, css in self.__pds.items():
            v = TimeImplictFunc(self.__create_closure(css, self.__gr.Sb_MVA))
            self.__gr.Bus(b).Pd += v
            if b in decs:
                assert isinstance(self.__sol.est, DistFlowSolver), "Load reduction only supported in DistFlowSolver"
                self.__sol.est.AddReduce(b, v)
        self.last_ok = GridSolveResult.Failed
        return 1e100

    def isSmartChargeEnabled(self)->bool:
        '''Check if smart charging is enabled'''
        if isinstance(self.__sol.est, DistFlowSolver):
            return len(self.__sol.est.DecBuses) > 0
        else:
            return False
    
    @property
    def Solver(self):
        return self.__sol
    
    @property
    def Grid(self)->Grid:
        '''Get the grid instance'''
        return self.__gr
    
    def PreStep(self, _t:int, /, sta:PluginStatus)->Tuple[bool,float]:
        '''Solve the optimal generation plan of the distribution network at time _t, 
        the solution result can be found in the relevant values of the Grid instance'''
        if sta == PluginStatus.EXECUTE:
            ok, val = self.__sol.solve(_t)
            if ok == GridSolveResult.Failed:
                print(f"[{_t}] Fail.", file = self.__fh)
                self.__gr.savePQofBus(os.path.join(self.__save_to, f"{_t}_load.csv"), _t)
                self.__badcnt += 1
                if self.__badcnt > 0 and self.__badcnt % 20 == 0:
                    print(_locale["PDN_SOLVE_FAILED_CNT"].format(self.__badcnt))
                if self.last_ok:
                    print(_locale["PDN_SOLVE_FAILED"].format(_t,self.__badcnt))
            else:
                self.__gr.ApplyAllESS(_t - self.LastTime)
                if ok == GridSolveResult.OKwithoutVICons or ok == GridSolveResult.SubOKwithoutVICons:
                    print(f"[{_t}] Relax.", file = self.__fh)
                if self.isSmartChargeEnabled():
                    assert isinstance(self.__sol.est, DistFlowSolver)
                    for c in chain(self.__inst.FCSList, self.__inst.SCSList):
                        c.set_Pc_lim(float("inf"))
                    for b,x in self.__sol.est.DecBuses.items():
                        if x.Reduction:
                            tot = x.Limit(_t)
                            k = (tot - x.Reduction) / tot
                            print(f"[{_t}] Load reduction at bus {b}:",
                                  f"{x.Reduction*self.__gr.Sb_kVA:.2f} kW", file = self.__fh)
                            for c in self.__pds[b]:
                                l = k * c.Pc
                                c.set_Pc_lim(l)
                                print(f"    CS {c.name}: {l*3600:.2f} kW <- {c.Pc_kW:.2f} kW", file = self.__fh)
            if ok != GridSolveResult.Failed and self.last_ok == GridSolveResult.Failed and self.__badcnt>0:
                print(_locale["PDN_SOLVE_OK_SINCE"].format(_t,self.__badcnt))
            self.last_ok = ok
            return ok != GridSolveResult.Failed, val
        elif sta == PluginStatus.OFFLINE:
            return True, 0
        elif sta == PluginStatus.HOLD:
            return self.LastPostStepSucceed, self.LastPreStepResult
    
    @property
    def BestCost(self)->float: 
        '''Get the cost of the best generation plan, requires the last solve to be successful'''
        assert self.LastPreStepSucceed, _locale["ERROR_LAST_FAILED"]
        return self.LastPreStepResult
    
    @property
    def GeneratorPlan(self)->Dict[str,float]:
        '''Get the best generation plan, requires the last solve to be successful'''
        assert self.LastPreStepSucceed, _locale["ERROR_LAST_FAILED"]
        return {g.ID: _sv(g.P) for g in self.__gr.Gens}


from v2simux_gui.common import *


class OptionBox(Frame):
    def __init__(self, master, options:Dict[str, Tuple[str, bool]], lcnt:int = -1, **kwargs):
        super().__init__(master, **kwargs)
        self._bools:List[BooleanVar] = []
        self._ctls:List[Checkbutton] = []
        self._mp:Dict[str, BooleanVar] = {}
        self._fr:List[Frame] = []
        if lcnt <= 0: 
            fr = Frame(self)
            fr.pack(side = "top", anchor = "w")
            self._fr.append(fr)
        i = 0
        for id, (text, v) in options.items():
            bv = BooleanVar(self, v)
            self._bools.append(bv)
            self._mp[id] = bv
            if lcnt > 0 and i % lcnt == 0:
                fr = Frame(self)
                fr.pack(side = "top", anchor = "w")
                self._fr.append(fr)
            self._ctls.append(Checkbutton(self._fr[-1],text=text,variable=bv))
            self._ctls[-1].pack(side='left',anchor="w")
            i+=1
    
    def disable(self):
        for c in self._ctls:
            c['state']=DISABLED
        
    def enable(self):
        for c in self._ctls:
            c['state']=NORMAL

    def __setitem__(self, key:str, value:bool):
        self._mp[key].set(value)
    
    def __getitem__(self, key:str)->bool:
        return self._mp[key].get()
    
    def getValues(self):
        return {k: v.get() for k,v in self._mp.items()}
    
    def getSelected(self):
        return [k for k,v in self._mp.items() if v.get()]


from v2simux_gui.common import *

from feasytools import RangeList, SegFunc

_L = LangLib.Load(__file__)
ALWAYS_ONLINE = _L['ALWAYS_ONLINE']

def empty_postfunc(itm:Tuple[Any,...], val:str): pass

PostFunc = Callable[[Tuple[Any,...], str], None]    
    
# Double click to edit the cell: https://blog.csdn.net/falwat/article/details/127494533
class ScrollableTreeView(Frame):
    def show_title(self, title:str):
        self.lb_title.config(text=title)
        self.lb_title.grid(row=0,column=0,padx=3,pady=3,sticky="w",columnspan=2)

    def hide_title(self):
        self.lb_title.grid_remove()

    def __init__(self, master, allowSave:bool = False, allowAdd:bool = False, allowDel:bool = False, 
                 allowMove:bool = False, addgetter:Optional[Callable[[], Optional[List[Any]]]] = None, **kwargs):
        super().__init__(master, **kwargs)
        self.post_func = empty_postfunc
        self._afterf = None
        self.lb_title = Label(self, text=_L["NOT_OPEN"])
        self.tree = Treeview(self)
        self.tree.grid(row=1,column=0,sticky='nsew')
        self.rowconfigure(1, weight=1)
        self.columnconfigure(0, weight=1)
        self.VScroll1 = Scrollbar(self, orient='vertical', command=self.tree.yview)
        self.VScroll1.grid(row=1, column=1, sticky='ns')
        self.HScroll1 = Scrollbar(self, orient='horizontal', command=self.tree.xview)
        self.HScroll1.grid(row=2, column=0, sticky='ew')
        self.tree.configure(yscrollcommand=self.VScroll1.set,xscrollcommand=self.HScroll1.set)
        self.bottom_panel = Frame(self)
        self.btn_save = Button(self.bottom_panel, text=_L["SAVE"], command=self.save)
        self.lb_save = Label(self.bottom_panel, text=_L["NOT_OPEN"])
        self.lb_note = Label(self.bottom_panel, text=_L["EDIT_NOTE"])
        self.btn_add = Button(self.bottom_panel, text=_L["ADD"], command=self.additm)
        self.btn_del = Button(self.bottom_panel, text=_L["DELETE"], command=self.delitm)
        self.btn_moveup = Button(self.bottom_panel, text=_L["UP"], command=self.moveup)
        self.btn_movedown = Button(self.bottom_panel, text=_L["DOWN"], command=self.movedown)
        self.addgetter = addgetter
        if allowSave or allowAdd or allowDel or allowMove:
            self.bottom_panel.grid(row=3,column=0,padx=3,pady=3,sticky="nsew")
        if allowSave:
            self.btn_save.grid(row=0,column=0,padx=3,pady=3,sticky="w")
            self.lb_save.grid(row=0,column=1,padx=3,pady=3,sticky="w")
            self.lb_note.grid(row=0,column=2,padx=20,pady=3,sticky="w")
        if allowAdd and self.addgetter is not None:
            self.btn_add.grid(row=0,column=3,pady=3,sticky="e")
        if allowDel:
            self.btn_del.grid(row=0,column=4,pady=3,sticky="e")
        if allowMove:
            self.btn_moveup.grid(row=0,column=5,pady=3,sticky="e")
            self.btn_movedown.grid(row=0,column=6,pady=3,sticky="e")
        self.delegate_var = StringVar()
        self.tree.bind('<Double-1>', func=self.tree_item_edit)
        self.onSave = None
        self.edit_mode:'Dict[str, Tuple[ConfigItem, PostFunc]]' = {}
        self.delegate_widget = None
        self.selected_item = None

    def additm(self):
        if self.addgetter:
            cols = self.addgetter()
            if cols is None or len(cols) != len(self.tree["columns"]):
                MB.showerror(_L["ERROR"], _L["ADD_FAILED"])
                return
            self.tree.insert("", "end", values=cols)
            self.lb_save.config(text=_L["UNSAVED"],foreground="red")
            if self._afterf: self._afterf()
    
    def delitm(self):
        dlist = [self.tree.item(x, "values")[0] for x in self.tree.selection()]
        if MB.askokcancel(_L["DELETE"], _L["DELETE_CONFIRM"].format(','.join(dlist))):
            for i in self.tree.selection():
                self.tree.delete(i)
            self.lb_save.config(text=_L["UNSAVED"],foreground="red")
            if self._afterf: self._afterf()

    def moveup(self):
        for i in self.tree.selection():
            p = self.tree.index(i)
            self.tree.move(i, "", p-1)
        self.lb_save.config(text=_L["UNSAVED"],foreground="red")
        if self._afterf: self._afterf()
    
    def movedown(self):
        for i in self.tree.selection():
            p = self.tree.index(i)
            self.tree.move(i, "", p+1)
        self.lb_save.config(text=_L["UNSAVED"],foreground="red")
        if self._afterf: self._afterf()
    
    def save(self):
        if self.onSave:
            if self.onSave(self.getAllData()):
                self.lb_save.config(text=_L["SAVED"],foreground="green")
    
    def setOnSave(self, onSave:Callable[[List[tuple]], bool]):
        self.onSave = onSave
    
    def item(self, item, option=None, **kw):
        return self.tree.item(item, option, **kw)
    
    def getAllData(self) -> List[tuple]:
        res = []
        for i in self.tree.get_children():
            res.append(self.tree.item(i, "values"))
        return res
    
    def setColEditMode(self, col:str, mode:ConfigItem, post_func:PostFunc = empty_postfunc):
        self.__setEditMode("COL:" + col, mode, post_func)

    def setRowEditMode(self, row:str, mode:ConfigItem, post_func:PostFunc = empty_postfunc):
        self.__setEditMode("ROW:" + row, mode, post_func)
        print(row,mode)

    def setCellEditMode(self, row:str, col:str, mode:ConfigItem, post_func:PostFunc = empty_postfunc):
        self.__setEditMode("CELL:" + row + "@" + col, mode, post_func)

    def clearEditModes(self):
        self.edit_mode.clear()
    
    def __setEditMode(self, label:str, cfgitm:ConfigItem, post_func:PostFunc = empty_postfunc):
        mode = cfgitm.editor
        if mode == EditMode.SPIN:
            if cfgitm.spin_range is None: cfgitm.spin_range = (0, 100)
        elif mode == EditMode.COMBO:
            if cfgitm.combo_values is None: cfgitm.combo_values = []
        self.edit_mode[label] = (cfgitm, post_func)

    def disableEdit(self):
        self.tree.unbind('<Double-1>')
    
    def enableEdit(self):
        self.tree.bind('<Double-1>', func=self.tree_item_edit)
    
    def tree_item_edit(self, e: Event):
        if len(self.tree.selection()) == 0:
            return
        
        self.selected_item = self.tree.selection()[0]
        selected_row = self.tree.item(self.selected_item, "values")[0]

        for i, col in enumerate(self.tree['columns']):
            x, y, w, h = self.tree.bbox(self.selected_item, col)
            assert isinstance(x, int) and isinstance(y, int) and isinstance(w, int) and isinstance(h, int)
            if x < e.x < x + w and y < e.y < y + h:
                self.selected_column = col
                text = self.tree.item(self.selected_item, 'values')[i]
                break
        else:
            self.selected_column = None
            x, y, w, h =  self.tree.bbox(self.selected_item)
            assert isinstance(x, int) and isinstance(y, int) and isinstance(w, int) and isinstance(h, int)
            text = self.tree.item(self.selected_item, 'text')
        
        self.delegate_var.set(text)
        possible_labels = []
        if self.selected_column is not None and selected_row is not None:
            possible_labels.append("CELL:" + selected_row + "@" + self.selected_column)
        if self.selected_column is not None:
            possible_labels.append("COL:" + self.selected_column)
        if selected_row is not None:
            possible_labels.append("ROW:" + selected_row)
        
        label = None
        for lb in possible_labels:
            if lb in self.edit_mode:
                label = lb
                break
        
        if label is None: return
        
        cfg, self.post_func = self.edit_mode[label]
        if cfg.editor == EditMode.COMBO:
            assert isinstance(cfg.combo_values, (list, tuple))
            self.delegate_widget = Combobox(self.tree, width=w // 10, textvariable=self.delegate_var, values=cfg.combo_values)
            self.delegate_widget.bind('<<ComboboxSelected>>', self.tree_item_edit_done)
            self.delegate_widget.bind('<FocusOut>', self.tree_item_edit_done)
        elif cfg.editor == EditMode.CHECKBOX:
            self.delegate_widget = Combobox(self.tree, width=w // 10, textvariable=self.delegate_var, values=[str(True), str(False)])
            self.delegate_widget.bind('<FocusOut>', self.tree_item_edit_done)
        elif cfg.editor == EditMode.SPIN:
            assert isinstance(cfg.spin_range, tuple) and len(cfg.spin_range) == 2
            self.delegate_widget = Spinbox(self.tree, width=w // 10, textvariable=self.delegate_var, from_=cfg.spin_range[0], to=cfg.spin_range[1], increment=1)
            self.delegate_widget.bind('<FocusOut>', self.tree_item_edit_done)
        elif cfg.editor == EditMode.ENTRY:
            self.delegate_widget = Entry(self.tree, width=w // 10, textvariable=self.delegate_var)
            self.delegate_widget.bind('<FocusOut>', self.tree_item_edit_done)
        elif cfg.editor == EditMode.RANGELIST:
            d = self.delegate_var.get()
            if d == ALWAYS_ONLINE: d = "[]"
            from .rle import RangeListEditor
            self.delegate_widget = RangeListEditor(RangeList(eval(d)), self.delegate_var, True)
            self.delegate_widget.bind('<Destroy>', self.tree_item_edit_done)
        elif cfg.editor == EditMode.PROP:
            assert isinstance(cfg.prop_config, ConfigItemDict)
            d = self.delegate_var.get()
            from .prope import PropertyEditor
            self.delegate_widget = PropertyEditor(eval(d), self.delegate_var, edit_modes=cfg.prop_config)
            self.delegate_widget.bind('<Destroy>', self.tree_item_edit_done)
        elif cfg.editor == EditMode.PDFUNC:
            from .pdfe import PDFuncEditor
            self.delegate_widget = PDFuncEditor(self.delegate_var)
            self.delegate_widget.bind('<Destroy>', self.tree_item_edit_done)
        elif cfg.editor == EditMode.SEGFUNC:
            d = self.delegate_var.get()
            obj = eval(d)
            if obj is None: obj = []
            elif isinstance(obj, (float, int)): obj = [(0,obj)]
            assert isinstance(obj, list)
            for xx in obj:
                assert isinstance(xx, tuple)
                assert isinstance(xx[0], int)
                assert isinstance(xx[1], (int,float))
            self.delegate_widget = SegFuncEditor(SegFunc(obj), self.delegate_var) # type: ignore
            self.delegate_widget.bind('<Destroy>', self.tree_item_edit_done)
        else:
            return
        if not isinstance(self.delegate_widget, Toplevel):
            self.delegate_widget.place(width=w, height=h, x=x, y=y)
        self.delegate_widget.focus()
        self.lb_save.config(text=_L["UNSAVED"],foreground="red")

    def tree_item_edit_done(self, e):
        if self.delegate_widget and not isinstance(self.delegate_widget, Toplevel):
            self.delegate_widget.place_forget()
        v = self.delegate_var.get()
        if not self.selected_item: return
        try:
            line = self.tree.item(self.selected_item, 'values')
        except:
            return
        assert isinstance(line, tuple)
        self.post_func(line, v)
        if self.selected_column is None:
            self.tree.item(self.selected_item, text=v)
        else:
            self.tree.set(self.selected_item, self.selected_column, v)
        if self._afterf: self._afterf()
    
    @property
    def AfterFunc(self):
        '''Function to be executed when an item is editted'''
        return self._afterf
    
    @AfterFunc.setter
    def AfterFunc(self, v):
        self._afterf = v
    
    def __setitem__(self, key, val):
        self.tree[key] = val
    
    def __getitem__(self, key):
        return self.tree[key]
    
    def column(self, *args, **kwargs):
        self.tree.column(*args, **kwargs)
    
    def heading(self, *args, **kwargs):
        self.tree.heading(*args, **kwargs)
    
    def insert(self, *args, **kwargs):
        self.tree.insert(*args, **kwargs)
    
    def delete(self, *items:Union[str,int]):
        self.tree.delete(*items)
    
    def set(self, item:Union[str, int], column:Union[None, str, int], value:Any):
        self.tree.set(item, column, value)
    
    def selection(self):
        return self.tree.selection()
    
    def get_children(self):
        return self.tree.get_children()
    
    def clear(self):
        self.delete(*self.get_children())
        self.lb_save.config(text=_L["SAVED"], foreground="green")
    
    @property
    def saved(self):
        return self.lb_save.cget("text") != _L["UNSAVED"]

__all__ = ["ScrollableTreeView", "ALWAYS_ONLINE", "empty_postfunc", "PostFunc"]